
/*
 * HPTalx
 * A PC<->HP48/49 communication program for Linux
 *
 * (C) 1999, 2000 by Bruno Barberi Gnecco
 * (C) 2005 by Rafael Ostertag
 *
 * $Id: hpcomm.c,v 1.4 2007/10/06 19:10:05 guengel Exp $
 *
 *
 * Copyright note:
 *
 * This program is free software; you can
 * redistribute it and/or modify it under the
 * terms of the GNU General Public License as
 * published by the Free Software Foundation;
 * either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that
 * it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU
 * General Public License along with this
 * program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "hpcomm.h"
#include "kermit.h"

#include <glib.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>

/*
 * Those are the characters that are interpreted by Kermit. Those are needed
 * by _hp_sanitize_input() to slashify 'em. 
 */
const char         *critical_chars[] = {
	"-",
	"\"",
	"<",
	">",
	NULL
};

void                _hp_replace_newline (char *src, size_t size);
unsigned short      _hp_convert_type_to_num (char *s);
void                _hp_sanitize_input (char **i);

/*--------------------------------------------------------------------------*/

/***********************************************************

 Function
 ========
 unsigned short _hp_convert_type_to_num (char *s);

 Input
 =====
 *s				pointer to a string describing the HP Data
				Type

 Output
 ======
 Return Value	The number of the HP Data Type as defined in
				hpcomm.h

 Description
 ===========
 The functino takes a string describing the HP Data Type and
 returns the corresponding Data Type Number as defined in
 hpcomm.h

 ***********************************************************/
unsigned short
_hp_convert_type_to_num (char *s)
{
	if (s == NULL)
		return HP_TYPE_UNKNOWN;

	if (strcasecmp (HP_STR_REAL, s) == 0)
		return HP_REAL;
	if (strcasecmp (HP_STR_COMPLEX, s) == 0)
		return HP_COMPLEX;
	if (strcasecmp (HP_STR_STRING, s) == 0)
		return HP_STRING;
	if (strcasecmp (HP_STR_REAL_ARRAY, s) == 0)
		return HP_REAL_ARRAY;
	if (strcasecmp (HP_STR_COMPLEX_ARRAY, s) == 0)
		return HP_COMPLEX_ARRAY;
	if (strcasecmp (HP_STR_LIST, s) == 0)
		return HP_LIST;
	if (strcasecmp (HP_STR_GLOBAL_NAME, s) == 0)
		return HP_GLOBAL_NAME;
	if (strcasecmp (HP_STR_LOCAL_NAME, s) == 0)
		return HP_LOCAL_NAME;
	if (strcasecmp (HP_STR_PROGRAM, s) == 0)
		return HP_PROGRAM;
	if (strcasecmp (HP_STR_ALGEBRAIC_OBJECT, s) == 0)
		return HP_ALG_OBJECT;
	if (strcasecmp (HP_STR_BINARY_INTEGER, s) == 0)
		return HP_BINT;
	if (strcasecmp (HP_STR_GRAPHICS_OBJECT, s) == 0)
		return HP_GRAPHICS_OBJECT;
	if (strcasecmp (HP_STR_TAGGED_OBJECT, s) == 0)
		return HP_TAGGED_OBJECT;
	if (strcasecmp (HP_STR_UNIT, s) == 0)
		return HP_UNIT_OBJECT;
	if (strcasecmp (HP_STR_XLIB, s) == 0)
		return HP_XLIB_NAME;
	if (strcasecmp (HP_STR_DIRECTORY, s) == 0)
		return HP_DIRECTORY;
	if (strcasecmp (HP_STR_LIBRARY, s) == 0)
		return HP_LIBRARY;

	/*
	 * There are more, but for now, this is sufficient
	 */
	return HP_TYPE_UNKNOWN;
}

/***********************************************************

 Function
 ========
 void _hp_replace_newline (char *src, size_t size);

 Input
 =====
 *src			pointer to a string
 size			size of the string

 Output
 ======
 NONE

 Description
 ===========
 Replaces all newline characters '\n' with a space ' ' in the
 string pointed to by *src.

 ***********************************************************/
void
_hp_replace_newline (char *src, size_t size)
{
	register size_t     i;

	if (src == NULL)
		return;

	for (i = 0; i <= size; i++) {
		if (src[i] == HP_CHAR_NEWLINE)
			src[i] = HP_CHAR_SPACE;
	}
}

/***********************************************************

 Function
 ========
 void _hp_sanitize_input(char **i);

 Input
 =====
 **i			string that needs to be sanitized.

 Output
 ======
 Return Value	NONE

 Description
 ===========
 Takes a pointer to a string pointer **i and puts a slash
 in front of each character from critical_chars[] found in
 *i.

 ***********************************************************/
void
_hp_sanitize_input (char **i)
{
	char               *tmp;
	char               *ptr;
	size_t              l;
	register int        c;

	if ((i == NULL) || (*i == NULL))
		return;

	/*
	 * Go thru each character specified in critical_chars[] 
	 */
	for (c = 0; critical_chars[c] != NULL; c++) {
		/*
		 * Pointer points to the start of the string 
		 */
		ptr = *i;
		do {
			/*
			 * Find from ptr the first occurence of critical_char[i] 
			 */
			ptr = strstr (ptr, critical_chars[c]);
			if (ptr != NULL) {
				/*
				 * Since we have found a critical character, resize the
				 * buffer to hold the slashified version of the char 
				 */
				*i = (char *) realloc (*i, strlen (*i) + 2);
				if (*i == NULL)
					return;
				/*
				 * Get the length from ptr till the end of *i +1 for the zero 
				 * terminator 
				 */
				l = strlen (ptr) + 1;
				/*
				 * Prepare the temporary buffer to hold the string starting
				 * with ptr till the end of *i 
				 */
				tmp = (char *) malloc (l);
				/*
				 * Copy over the remainder of the string from position ptr 
				 */
				memcpy ((void *) tmp, (void *) ptr, l);
				/*
				 * Copy the slashes to the position of ptr 
				 */
				memcpy ((void *) ptr, (void *) HP_STR_BACKSLASH,
						strlen (HP_STR_BACKSLASH));
				/*
				 * Now increase the ptr by one, so it points after the slash 
				 */
				ptr++;
				/*
				 * Copy the remainder copied into tmp back to *i at position
				 * ptr 
				 */
				memcpy ((void *) ptr, (void *) tmp, l);
				/*
				 * Get rid of the buffer 
				 */
				free (tmp);
				/*
				 * Advance ptr so it does not point any more to the critical
				 * character and we won't get into an endless-loop 
				 */
				ptr++;
			}
		}
		while (ptr != NULL);
	}
}

/***********************************************************

 Function
 ========
 hp_status hp_version (hp_calc * v);

 Input
 =====
 NONE

 Output
 ======
 *v				pointer to a hp_calc enum.
 Return Value	return the status of the call

 Description
 ===========
 The function tries to retrieve the version of the HP
 Calculator connected and stores it in *v (see hpcomm.h
 enum _hp_calc for possible values).

 It returns HP_OK on success. On a Kermit error HP_KERMIT.
 The Kermit error code is stored in k_errno. For other
 errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_version (hp_calc * v)
{
	k_status            kstat;
	char               *bulk_data;
	size_t              wdc;

	K_CHECKSTATE;

	kstat = k_send_remote_command (HP_CMD_3TRANSIO);
	HP_CHECK_KSTAT (kstat);

	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	kstat = k_send_remote_command (HP_CMD_CLEAR);
	HP_CHECK_KSTAT (kstat);

	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	kstat = k_send_remote_command (HP_CMD_VERSION);
	HP_CHECK_KSTAT (kstat);

	bulk_data = NULL;
	kstat = k_wait_ready (&bulk_data, &wdc);
	if (kstat != K_OK) {
		k_free_bulk_output (&bulk_data);
		return HP_KERMIT;
	}

	if (strstr (bulk_data, HP_CALC_MODEL_48))
		*v = HP48;
	if (strstr (bulk_data, HP_CALC_MODEL_49GPLUS))
		*v = HP49G_PLUS;
	if (strstr (bulk_data, HP_CALC_MODEL_50G))
		*v = HP50G;
	/*
	 * Keep this here, else we can't decide whether it is an hp49gplus or just
	 * an hp49.
	 */
	if (strstr (bulk_data, HP_CALC_MODEL_49))
		*v = HP49G;

	k_free_bulk_output (&bulk_data);

	kstat = k_send_remote_command (HP_CMD_CLEAR);
	HP_CHECK_KSTAT (kstat);

	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_get_directory (const char *dn, GList ** list);

 Input
 =====
 *dn			pointer to a HP Directory String.
				(Format: "{ HOME xxx xxx }")

 Output
 ======
 **list			pointer to a pointer to a GList
 Return Value	returns the status of the call

 Description
 ===========
 The function takes in *dn a HP Directory String in the form
 { HOME xxx xxx } and returns a list of the entries in that
 HP Directory in a GList. For each entry in the HP Directory
 a list item holds a pointer to a hp_directory_entry
 structure declared in hpcomm.h.

 The caller is responsible for freeing the GList list and
 all associated hp_directory_entry structs.

 The function returns HP_OK on success. On a Kermit error
 HP_KERMIT.  The Kermit error code is stored in k_errno.
 For other errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_get_directory (const char *dn, GList ** list)
{
	hp_directory_entry *hp_dirent;
	k_status            kstat;
	hp_status           hpstat;
	char               *bulk_data = NULL;
	char               *data_copy;
	char               *bak_ptr;
	char               *ptr;
	char                hp_dirname[HP_DIRNAMESIZE];
	char                hp_typename[HP_TYPENAMESIZE];
	float               wdc;
	float               size;
	size_t              size_data;

	if ((dn == NULL) || (list == NULL) || (*list != NULL))
		return HP_INVARG;

	K_CHECKSTATE;

	/*
	 * Change to the directory 
	 */
	hpstat = hp_chdir (dn);
	if (hpstat != HP_OK)
		return hpstat;

	/*
	 * Retrieve directory listing 
	 */
	kstat = k_send_command (K_CMD_REMOTE_DIRECTORY);
	HP_CHECK_KSTAT (kstat);

	kstat = k_wait_ready (&bulk_data, &size_data);
	if (kstat != K_OK) {
		k_free_bulk_output (&bulk_data);
		return HP_KERMIT;
	}

	/*
	 * Prepare for parsing output
	 */

	data_copy = (char *) malloc (size_data);
	if (data_copy == NULL) {
		k_free_bulk_output (&bulk_data);
		return HP_MALLOC;
	}
	memcpy ((void *) data_copy, (void *) bulk_data, size_data);

	/*
	 * Backup Pointer
	 */
	bak_ptr = data_copy;

	/*
	 * Discard the first entry, since that is the pwd. Try it the easy way,
	 * remove the everthing in front of the first newline character 
	 */

  REDOFROMSTART:
	ptr = strsep (&data_copy, HP_STR_NEWLINE);
	if (data_copy == NULL) {
		free (ptr);
		k_free_bulk_output (&bulk_data);
		return HP_PARSE;
	}

	/*
	 * Check if there are any current pwd leftovers 
	 */
	if (strstr (data_copy, HP_STR_DIR_END) != NULL) {
		/*
		 * Hell, there are, get rid of 'em 
		 */
		ptr = strsep (&data_copy, HP_STR_DIR_END);
		if (data_copy == NULL) {
			/*
			 * WTF are we dealing with? Bail out 
			 */
			free (ptr);
			k_free_bulk_output (&bulk_data);
			return HP_PARSE;
		}
		/*
		 * And remove again the newline character 
		 */
		goto REDOFROMSTART;
	}

	do {
		/*
		 * This are the entries we are interrested in
		 */
		ptr = strsep (&data_copy, HP_STR_NEWLINE);
		if (data_copy == NULL)
			break;

		/*
		 * Set buffers to zero
		 */
		memset ((void *) hp_dirname, 0, HP_DIRNAMESIZE);
		memset ((void *) hp_typename, 0, HP_TYPENAMESIZE);
		sscanf (ptr, "%s%f%s%f", hp_dirname, &size, hp_typename, &wdc);

		hp_dirent = (hp_directory_entry *) malloc (sizeof (hp_dirent) + 1);
		hp_dirent->name = strdup (hp_dirname);
		hp_dirent->type = _hp_convert_type_to_num (hp_typename);
		hp_dirent->size = (long int) size;

		*list = g_list_append (*list, (gpointer) hp_dirent);

	}
	while (data_copy != NULL);
	data_copy = bak_ptr;

	free (data_copy);
	k_free_bulk_output (&bulk_data);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_get_current_directory(char *dir, size_t s);

 Input
 =====
 s				size of the buffer *dir

 Output
 ======
 *dir			buffer to hold the current directory.

 Return Value	status of the call

 Description
 ===========
 The function retrieves the current directory of the HP
 Calculator into *dir. s specifies the size of *dir.

 The function returns HP_OK on success. On a Kermit error
 HP_KERMIT. The Kermit error code is stored in k_errno.
 For other  errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_get_current_directory (char *dir, size_t s)
{
	k_status            kstat;
	char               *bulk_data;
	size_t              data_size;
	char               *dir_start = NULL;
	char               *dir_end = NULL;
	register size_t     i;

	K_CHECKSTATE;

	/*
	 * Zero out things, so if we fail, the user does not get garbage 
	 */
	memset (dir, 0, s);

	/*
	 * We cannot use the RPL Command "PATH" since its output gets truncated.
	 * This appears to be new with at least HP49G+ since the HP48GX does not
	 * show any problems with long paths. 
	 */
	kstat = k_send_command (K_CMD_REMOTE_DIRECTORY);
	HP_CHECK_KSTAT (kstat);

	bulk_data = NULL;
	kstat = k_wait_ready (&bulk_data, &data_size);
	if (kstat != K_OK) {
		k_free_bulk_output (&bulk_data);
		return HP_KERMIT;
	}

	for (i = 0; i < data_size; i++) {
		/*
		 * Try to find the start of the HP Directory String 
		 */
		if (bulk_data[i] == HP_CHAR_DIR_START)
			dir_start = bulk_data + i;
		/*
		 * Try to find the end of the HP Directory String 
		 */
		if (bulk_data[i] == HP_CHAR_DIR_END) {
			dir_end = bulk_data + i + 1;
			/*
			 * No need to search any further 
			 */
			break;
		}
	}

	/*
	 * Check if the buffer is large enough to hold the HP Directory String 
	 */
	if (s < (dir_end - dir_start)) {
		k_free_bulk_output (&bulk_data);
		return HP_BUFFER;
	}
	/*
	 * Copy over the HP Directory String 
	 */
	strncpy (dir, dir_start, (dir_end - dir_start));

	/*
	 * Since there could be newlines, replace 'em with spaces 
	 */
	for (i = 0; i < (dir_end - dir_start); i++) {
		if (dir[i] == HP_CHAR_NEWLINE) {
			dir[i] = HP_CHAR_SPACE;
		}
	}

	k_free_bulk_output (&bulk_data);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_chdir(const char *hpdir);

 Input
 =====
 *hpdir			pointer to an existing string holding HP
				Directory String.
				(Format: "{ HOME xxx xxx }")

 Output
 ======
 Return Value	status of the call

 Description
 ===========
 The function changes the directory on the HP Calculator to
 *hpdir. *hpdir must be a HP Directory in the form:
 "{ HOME xxx xxx }".

 On success, it returns HP_OK. On a Kermit error HP_KERMIT.
 The Kermit error code is stored in k_errno. For other
 errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_chdir (const char *hpdir)
{
	k_status            kstat;

	if (hpdir == NULL)
		return HP_INVARG;

	kstat = k_send_remote_command (HP_CMD_SINGLE_EVAL, hpdir);
	HP_CHECK_KSTAT (kstat);

	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status	hp_mkdir(const char *dir);

 Input
 =====
 *dir			pointer to an existing string holding the
				directory name.

 Output
 ======
 Return Value	status of the call

 Description
 ===========
 The function creates a new directory *dir in the current
 working directory. *dir must not be a HP Directory String.

 On success, it returns HP_OK. On a Kermit error HP_KERMIT.
 The Kermit error code is stored in k_errno. For other
 errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_mkdir (const char *dir)
{
	k_status            kstat;

	if (dir == NULL)
		return HP_INVARG;

	kstat = k_send_remote_command (HP_CMD_MKDIR, dir);
	HP_CHECK_KSTAT (kstat);

	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_directory_push (const char *hpdir,
							  const char *d,
							  char *newdir,
							  size_t s);

 Input
 =====
 *hpdir			pointer to an existing string holding HP
				Directory String.
				(Format: "{ HOME xxx xxx }")

 *d				pointer to an existing string holding the
				directory pushed to *hpdir

 s				size of *newdir in bytes

 Output
 ======
 *newdir		pointer to a buffer that will store the
				new HP Directory String

 Return Value	status of the call

 Description
 ===========
 The function adds a new directory *d to an existing HP
 Directory String *hpdir. The new HP Directory String is
 returned in *newdir

 The function returns HP_OK on success. On error, it returns
 *hpdir in *newdir and sets the return value accordingly.
 For possible error codes, see the hp_status enum in
 hpcomm.h.

 ***********************************************************/
hp_status
hp_directory_push (const char *hpdir, const char *d, char *newdir, size_t s)
{
	char               *end_dir;

	if (hpdir == NULL || d == NULL || newdir == NULL || s == 0)
		return HP_INVARG;

	/*
	 * If there is not enought space, simply return hpdir in newdir
	 */
	if (strlen (hpdir) > s) {
		strncpy (newdir, hpdir, s - 1);
		return HP_BUFFER;
	}

	/*
	 * Get closing curly brace
	 */
	end_dir = strstr (hpdir, HP_STR_DIR_END);

	/*
	 * Check whether closing curly brace is in
	 * place
	 */
	if (end_dir == NULL)
		return HP_OK;

	/*
	 * Is there a space in front of the closing
	 * curly brace?
	 */
	if (hpdir[(end_dir - hpdir) - 1] == HP_CHAR_SPACE) {
		/*
		 * We jump to the space
		 */
		end_dir--;
	}

	memset (newdir, 0, s);
	/*
	 * Copy hpdir without curly brace over to
	 * newdir.
	 */
	strncpy (newdir, hpdir, end_dir - hpdir);
	/*
	 * Add space 
	 */
	strncat (newdir, HP_STR_SPACE, 1);

	/*
	 * Check if there is enought space
	 */
	if (strlen (newdir) + strlen (d) + 2 > s) {
		/*
		 * Return hpdir in newdir
		 */
		strncpy (newdir, hpdir, strlen (hpdir) + 1);
		return HP_BUFFER;
	}
	/*
	 * Add new directory
	 */
	strncat (newdir, d, strlen (d));
	strncat (newdir, HP_STR_DIR_END, 1);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_directory_pop (const char *hpdir,
							 char *newdir,
							 size_t s);

 Input
 =====
 *hpdir			pointer to an existing HP Directory String.
				(Format: "{ HOME xxx xxx }")

 s				size of *newdir in bytes.

 Output
 ======
 *newdir		pointer to a buffer that will store the new
				HP Directory String

 Return Value	status of the function call

 Description
 ===========
 The function pops the last directory found in the
 HP Directory String *hpdir and returns the resulting
 HP Directory String in *newdir.

 The function returns HP_OK on success. On error it returns
 *hpdir in *newdir and sets the return value accordingly.
 For possible error codes, see the hp_status enum in
 hpcomm.h.

 ***********************************************************/
hp_status
hp_directory_pop (const char *hpdir, char *newdir, size_t s)
{
	register int        i;

	if (hpdir == NULL || newdir == NULL || s == 0)
		return HP_INVARG;

	for (i = strlen (hpdir); i >= 0; i--) {
		if (hpdir[i] == HP_CHAR_DIR_END) {
			/*
			 * Skipp the leading blank in front
			 * of the closing curly brace
			 */
			if (hpdir[i - 1] == HP_CHAR_SPACE) {
				i--;
				continue;
			}
		}
		/*
		 * Search for the next space
		 */
		if (hpdir[i] == HP_CHAR_SPACE) {
			/*
			 * We have found it
			 */
			break;
		}
	}

	/*
	 * i now marks the second last entry in the
	 * directory list
	 */
	memset (newdir, 0, s);

	/*
	 * Check if there is enought space
	 */
	if (i + 2 > s) {
		strncpy (newdir, hpdir, s - 1);
		return HP_BUFFER;
	}

	/*
	 * Copy from hpdir to newdir ommiting the
	 * last entry
	 */
	strncpy (newdir, hpdir, i);
	strncat (newdir, HP_STR_DIR_END, 1);

	return HP_OK;

}

/***********************************************************

 Function
 ========
 hp_status hp_directory_unix_style (char *hpdir,
									char *ustyle,
								    size_t s);

 Input
 =====
 *hpdir			pointer to a HP Directory String.
				(Format: "{ HOME xxx xxx }")

 s				size of the buffer holding the converted
				string

 Output
 ======
 *ustyle		pointer to a buffer that will store the
				converted string.

 Return Value	status of the function call

 Description
 ===========
 The function takes a HP Directory String in *hpdir and
 replaces all spaces with forward slashes '/' and stores the
 new string in *ustyle.

 The function returns HP_OK on success. If the size s of
 the buffer *ustyle is too small to hold the new string,
 HP_BUFF is returned. For other possible error codes, see
 the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_directory_unix_style (char *hpdir, char *ustyle, size_t s)
{
	register int        i;
	char               *dir_start;
	char               *dir_end;

	if (hpdir == NULL || ustyle == NULL || s == 0)
		return HP_INVARG;

	/*
	 * Make sure we have a hp directory in the form
	 * "{HOME xxx xxx xxx}". If not, return hpdir in ustyle.
	 */
	if (strstr (hpdir, HP_STR_DIR_START) != NULL
		&& strstr (hpdir, HP_STR_SLASH) == NULL) {
		strncpy (ustyle, hpdir, s);
		return HP_OK;
	}

	for (i = 0; i < strlen (hpdir); i++) {
		/*
		 * Pin down start of directory names
		 */
		if (hpdir[i] == HP_CHAR_DIR_START) {
			/*
			 * If there is a blank after the curly brace, ommit it
			 */
			if (hpdir[i + 1] == HP_CHAR_SPACE) {
				dir_start = hpdir + i + 2;
			} else {
				dir_start = hpdir + i + 1;
			}
		}
		/*
		 * Pin down end of directory names
		 */
		if (hpdir[i] == HP_CHAR_DIR_END) {
			/*
			 * If there is a blank in front of the curly brace, ommit it.
			 */
			if (hpdir[i - 1] == HP_CHAR_SPACE) {
				dir_end = hpdir + (i - 2);
			} else {
				dir_end = hpdir + (i - 1);
			}
		}
	}

	memset ((void *) ustyle, 0, s);

	/*
	 * Check if there is enough space in ustyle according to s.
	 * +3 because 1 is for the leading /, 1 is for the trailing /
	 * and 1 has to left for the terminating zero
	 */
	if (s < (dir_end - dir_start + 3))
		return HP_BUFFER;

	/*
	 * Now we copy over the directory names to ustyle
	 */
	strncpy (ustyle, HP_STR_SLASH, strlen (HP_STR_SLASH));
	strncat (ustyle, dir_start, (dir_end - dir_start) + 1);
	strncat (ustyle, HP_STR_SLASH, strlen (HP_STR_SLASH));

	/*
	 * THIS IS A HACK!
	 * We replace all spaces with /
	 */
	for (i = 0; i < strlen (ustyle); i++) {
		if (ustyle[i] == HP_CHAR_SPACE) {
			ustyle[i] = HP_CHAR_SLASH;
		}
	}

	/*
	 * There is no need to zero terminate string, because we zeroed out the entire buffer
	 */
	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_get_file_version (const char *f,
								hp_file_type * v)

 Input
 =====
 *f				string holding a Unix file path

 Output
 ======
 *v				pointer to a hp_file_type enum
 Return Value	status of the function call

 Description
 ===========
 The function opens the file *f and reads the first
 HP_FILEHEADERSIZE bytes of that file in order to figure out
 the HP File Type. Depending on the outcome, it sets *v to
 the following values:

 Value *v				Meaning
 -----------------------------------------------------------
 HP_FT_ANY				HP Text File. File that safely can
						be stored on any HP48/HP49
						Calculator.

 HP_FT_HP48				HP Binary File. File that should
						only be stored on a HP48 Calculator.

 HP_FT_HP49				HP Binary File. File that should
						only be stored on a HP489 Calculator.

 HP_FT_UNKNOWN			The file type could not be
						determined.

 For possible error codes, see the hp_status enum in
 hpcomm.h.

 ***********************************************************/
hp_status
hp_get_file_version (const char *f, hp_file_type * v)
{
	int                 fd;
	ssize_t             ret_val;
	char                buff[HP_FILEHEADERSIZE];

	if (f == NULL || v == NULL)
		return HP_INVARG;

	fd = open (f, O_RDONLY);
	if (fd < 0) {
		return HP_FOPEN;
	}

	ret_val = read (fd, (void *) buff, HP_FILEHEADERSIZE - 1);
	if (ret_val < 0) {
		close (fd);
		return HP_FREAD;
	}
	buff[ret_val] = 0;
	close (fd);

	*v = HP_FT_UNKNOWN;
	if (strstr (buff, HP_FT_STR_HP49)) {
		*v = HP_FT_HP49;
	}

	if (strstr (buff, HP_FT_STR_HP48)) {
		*v = HP_FT_HP48;
	}

	if (strstr (buff, HP_FT_STR_ANY)) {
		*v = HP_FT_ANY;
	}

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_rename (const char* oldname,
	 			      const char* newname,
					  const short int isdirectory);

 Input
 =====
 *oldname		name of the old directory/variable
 *newname		name of the new directory/variable
 isdirectory	flag indicating whether a directory or
				variable should be renamed.

 Output
 ======
 Return Value	status of the function call

 Description
 ===========
 The function takes the old name in *oldname and the new
 name in *newname. Depending on the flag isdirectory it
 handles *oldname as directory (isdirectory==HP_REN_DIR)
 or as variable (isdirectory==HP_REN_VAR). The function
 operates on the current directory on the HP Calculator.

 On success it returns HP_OK. On a Kermit error HP_KERMIT.
 The Kermit error code is stored in k_errno. For other
 errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_rename (const char *oldname, const char *newname,
		   const hp_rename_type isdirectory)
{

	k_status            kstat;

	K_CHECKSTATE;

	if (oldname == NULL || newname == NULL)
		return HP_INVARG;

	if (isdirectory == HP_REN_DIR)
		kstat = k_send_remote_command ("'%s' RCL '%s' STO '%s' PGDIR",
									   oldname, newname, oldname);
	else
		kstat = k_send_remote_command ("'%s' RCL '%s' STO '%s' PURGE",
									   oldname, newname, oldname);
	HP_CHECK_KSTAT (kstat);

	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_purge (const char *name,
					 const hp_purge_type isdirectory);

 Input
 =====
 *name			name of the entry to purge.
 isdirectory	flag indicating whether a directory or
				variable should be renamed.

 Output
 ======
 Return Value	status of the function call

 Description
 ===========
 The function purges the entry specified by *name in the
 current working directory. Depending on the flag
 isdirectory, it expects *name to be a directory
 (isdirectory==HP_PURGE_DIR) or a variable (isdirectory==
 HP_PURGE_VAR). Please note: If purging a directory, *name
 must not be a HP Directory String, though.

 On success it returns HP_OK. On a Kermit error HP_KERMIT.
 The Kermit error code is stored in k_errno. For other
 errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_purge (const char *name, const hp_purge_type isdirectory)
{
	k_status            kstat;

	if (name == NULL)
		return HP_INVARG;

	if (isdirectory == HP_PURGE_DIR)
		kstat = k_send_remote_command (HP_CMD_PURGE_DIR, name);
	else
		kstat = k_send_remote_command (HP_CMD_PURGE_VAR, name);

	HP_CHECK_KSTAT (kstat);

	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_send_file (const char *filename,
						 const hp_file_type ft,
						 k_transfer_callback callback,
						 void *cb_data);

 Input
 =====
 *filename		the file to be sent to the HP Calculator
 ft				the HP File Type of *filename
 callback		pointer to the function that is passed
				k_read_file_transfer()
 *cb_data		data which is passed to callback.

 Output
 ======
 Return Value	status of the function call

 Description
 ===========
 The function takes the file name to transfer to the HP
 Calculator in *filename and the type of the file in ft.

 It then sends the file to the HP Calculator and calls
 k_read_file_transfer() for monitoring the transfer, passing
 callback and *cb_data to it.

 On success it returns HP_OK. On a Kermit error HP_KERMIT.
 The Kermit error code is stored in k_errno. For other
 errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_send_file (const char *filename, const hp_file_type ft,
			  k_transfer_callback callback, void *cb_data)
{
	k_status            kstat;

	K_CHECKSTATE;

	if (filename == NULL)
		return HP_INVARG;

	if (ft == HP_FT_ANY) {
		/*
		 * It's probably a text file
		 */
		kstat = k_send_command (K_CMD_SEND_SINGLE_FILE_TEXT, filename);
	} else {
		/*
		 * Probably a binary file 
		 */
		kstat = k_send_command (K_CMD_SEND_SINGLE_FILE_BINARY, filename);
	}
	HP_CHECK_KSTAT (kstat);

	kstat = k_read_file_transfer (callback, cb_data);
	HP_CHECK_KSTAT (kstat);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_get_file (const char *name,
						k_transfer_type tt,
						k_transfer_callback callback,
						void *cb_data);

 Input
 =====
 *name			the file name of the file to get from the
				HP Calculator.
 tt				the transfer type
 callback		pointer to the function that is passed
				k_read_file_transfer()
 *cb_data		data which is passed to callback.

 Output
 ======
 Return Value	status of the function call

 Description
 ===========
 The function copies the file specified by *name from the
 current directory and initiates a receiving of the file
 to the PC. tt specifies the transfer mode (see
 k_transfer_type in kermit.h). callback is passed to
 k_read_file_transfer() to monitor the transfer. *cb_data is
 the argument to callback.

 The function returns HP_OK on success. On a Kermit error
 HP_KERMIT. The Kermit error code is stored in k_errno.
 For other errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_get_file (const char *name, k_transfer_type tt,
			 k_transfer_callback callback, void *cb_data)
{
	k_status            kstat;

	K_CHECKSTATE;

	if (name == NULL)
		return HP_INVARG;

	kstat = k_send_command ("get %s %s", k_transfer_file_switches[tt], name);
	HP_CHECK_KSTAT (kstat);

	kstat = k_read_file_transfer (callback, cb_data);
	HP_CHECK_KSTAT (kstat);

	return HP_OK;

}

/***********************************************************

 Function
 ========
 hp_status hp_copy_file(const char* name,
						const char *dest);

 Input
 =====
 *name			the name of the variable or directory to
				copied over to *dest
 *dest			the destination directory as a HP
				Directory String. (Format: "{ HOME xxx xxx}")

 Output
 ======
 Return Value	status of the function call

 Description
 ===========
 The function copies *name from the current HP Directory to
 *dest

 The function returns HP_OK on success. On a Kermit error
 HP_KERMIT. The Kermit error code is stored in k_errno.
 For other errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_copy_file (const char *name, const char *dest)
{
	k_status            kstat;
	hp_status           hpstat;
	char                curr_hp_dir[HP_DIRNAMESIZE];

	K_CHECKSTATE;

	if ((name == NULL) || (dest == NULL))
		return HP_INVARG;

	/*
	 * Safe the current HP Directory 
	 */
	hpstat = hp_get_current_directory (curr_hp_dir, HP_DIRNAMESIZE);
	HP_CHECK_HPSTAT (hpstat);

	/*
	 * What it does: "'%s' RCL": Recall the variable/directory *name. "%s EVAL":
	 * Change to the directory *dest. "'%s' STO": Store the data on the stack
	 * into *name in the new directory. "%s EVAL": switch back to the current
	 * directory *curr_hp_dir.
	 */
	kstat =
		k_send_remote_command ("'%s' RCL %s EVAL '%s' STO %s EVAL",
							   name, dest, name, curr_hp_dir);
	HP_CHECK_KSTAT (kstat);

	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_command (const char *c, char **output);

 Input
 =====
 *c				command as string.

 Output
 ======
 **output		buffer where to output will be stored in
 Return Value	status of the function call

 Description
 ===========
 The function executes the command *c on the HP Calculator
 and returns the output of it in **output. The caller is
 responsible for freeing the memory occupied by **output.

 The function returns HP_OK on success. On a Kermit error
 HP_KERMIT. The Kermit error code is stored in k_errno.
 For other errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_command (const char *command, char **output)
{
	k_status            kstat;
	char               *c;
	char               *kout;
	char                stack_number[HP_STACKNUMBERSIZE];
	char               *line_begin;
	char               *line_end;
	char               *line;
	GList              *stack = NULL;
	GList              *stack_start;
	size_t              output_size;
	size_t              stack_size;
	register int        i;

	K_CHECKSTATE;

	if ((command == NULL) || (output == NULL))
		return HP_INVARG;

	/*
	 * Slashify all critical characters 
	 */
	c = strdup (command);
	_hp_sanitize_input (&c);

	kstat = k_send_remote_command ("%s", c);
	if (kstat != K_OK) {
		free (c);
	}

	kout = NULL;
	kstat = k_wait_ready (&kout, &output_size);
	if (kstat != K_OK) {
		free (c);
		k_free_bulk_output (&kout);
		return HP_KERMIT;
	}

	/*
	 * Here we go through the output of k_wait_ready() and search for stack
	 * numbers, beginning with "1:" 
	 */
	for (i = 1; i <= HP_MAXSTACKNUMBER; i++) {
		/*
		 * Generate the stack number 
		 */
		snprintf (stack_number, HP_STACKNUMBERSIZE, "%i:", i);
		/*
		 * Search for this line number 
		 */
		line_begin = strstr (kout, stack_number);

		/*
		 * Check if there is no first line, hence the stack on HP Calc is
		 * empty 
		 */
		if (line_begin == NULL && i == 1) {
			/*
			 * There is no stack returned 
			 */
			line = strdup ("No Result");
			stack = g_list_prepend (stack, line);
			break;
		}

		/*
		 * Check if we found no further lines 
		 */
		if (line_begin == NULL)
			break;

		/*
		 * Ok, line_begin is not NULL, so let's search for the end of the
		 * line 
		 */
		line_end = strstr (line_begin, HP_STR_NEWLINE);
		if (line_end == NULL) {
			/*
			 * This is strange, so we bail out 
			 */
			break;
		}

		/*
		 * Create buffer to hold line 
		 */
		line = (char *) malloc (line_end - line_begin + 2);
		if (line == NULL)
			goto THEEND;

		/*
		 * Zero out everything, so we don't have to care about zero
		 * termination 
		 */
		memset (line, 0, (line_end - line_begin + 2));

		/*
		 * Copy the line in our buffer 
		 */
		memcpy (line, line_begin, (line_end - line_begin) + 1);

		/*
		 * Add this line to the glist 
		 */
		stack = g_list_append (stack, line);
	}

	stack = g_list_reverse (stack);
	/*
	 * Now we copy the GList into the buffer we return to the caller 
	 */
	stack_start = stack = g_list_first (stack);
	stack_size = 0;
	*output = NULL;
	while (stack) {
		stack_size += strlen ((char *) stack->data);
		if (*output == NULL) {
			*output = (char *) malloc (stack_size);
			if (*output == NULL)
				goto THEEND;
			memcpy (*output, stack->data, strlen ((char *) stack->data));
		} else {
			*output = (char *) realloc (*output, stack_size);
			if (*output == NULL)
				goto THEEND;
			memcpy ((*output + (stack_size - strlen ((char *) stack->data))),
					stack->data, strlen ((char *) stack->data));
		}
		stack = g_list_next (stack);
	}
	/*
	 * Zero terminate buffer 
	 */
	stack_size++;
	*output = (char *) realloc (*output, stack_size);
	if (*output == NULL)
		goto THEEND;

	(*output)[stack_size - 1] = HP_CHAR_NULL;

  THEEND:
	/*
	 * Free GList 
	 */
	stack = stack_start;
	while (stack) {
		free (stack->data);
		stack = g_list_next (stack);
	}
	g_list_free (stack_start);
	k_free_bulk_output (&kout);
	free (c);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_store_flags ();

 Input
 =====
 NONE

 Output
 ======
 Return Value	status of the function call

 Description
 ===========
 The function stores the HP Calculator flags in a variable
 named as defined in HP_FLAG_VARIABLE_NAME and puts the
 program to restore the flags in a variable named as defined
 in HP_FLAG_RESTORE_NAME. But variables are stored in the
 { HOME } directory on the HP Calculator

 The function returns HP_OK on success. On a Kermit error
 HP_KERMIT. The Kermit error code is stored in k_errno.
 For other errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_store_flags ()
{
	k_status            kstat;
	hp_status           hpstat;
	char                currdir[HP_DIRNAMESIZE];

	/*
	 * Retrieve the current directory on the HP Calculator 
	 */
	hpstat = hp_get_current_directory (currdir, HP_DIRNAMESIZE);
	HP_CHECK_HPSTAT (hpstat);

	/*
	 * Change to the HP Calculator Home Directory 
	 */
	hpstat = hp_chdir (HP_HOME);
	HP_CHECK_HPSTAT (hpstat);

	/*
	 * Recall the flags 
	 */
	kstat = k_send_remote_command (HP_CMD_RECALL_FLAGS);
	HP_CHECK_KSTAT (kstat);
	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	/*
	 * Store the recalled flags 
	 */
	kstat = k_send_remote_command ("'%s' STO", HP_FLAG_VARIABLE_NAME);
	HP_CHECK_KSTAT (kstat);
	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	/*
	 * Create the program to recall the flags and store it into a variable
	 * named as defined in HP_FLAG_RESTORE_NAME 
	 */
	kstat =
		k_send_remote_command
		("\\\\<< '%s' RCL STOF { %s %s } PURGE",
		 HP_FLAG_VARIABLE_NAME, HP_FLAG_VARIABLE_NAME, HP_FLAG_RESTORE_NAME);
	HP_CHECK_KSTAT (kstat);
	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);

	kstat = k_send_remote_command ("'%s' STO", HP_FLAG_RESTORE_NAME);
	HP_CHECK_KSTAT (kstat);
	kstat = k_discard_output ();
	HP_CHECK_KSTAT (kstat);
	/*
	 * Change to the saved working directory 
	 */
	hpstat = hp_chdir (currdir);
	HP_CHECK_HPSTAT (hpstat);

	return HP_OK;
}

/***********************************************************

 Function
 ========
 hp_status hp_purge_stored_flags ();

 Input
 =====
 NONE

 Output
 ======
 Return Value	status of the function call

 Description
 ===========
 The function removes both variables created by a former
 call to hp_store_flags().

 The function returns HP_OK on success. On a Kermit error
 HP_KERMIT. The Kermit error code is stored in k_errno.
 For other errors, see the hp_status enum in hpcomm.h.

 ***********************************************************/
hp_status
hp_purge_stored_flags ()
{
	hp_status           hpstat;
	char                currdir[HP_DIRNAMESIZE];

	/*
	 * Retrieve the current directory on the HP Calculator 
	 */
	hpstat = hp_get_current_directory (currdir, HP_DIRNAMESIZE);
	HP_CHECK_HPSTAT (hpstat);

	/*
	 * Change to the HP Calculator Home Directory 
	 */
	hpstat = hp_chdir (HP_HOME);
	HP_CHECK_HPSTAT (hpstat);

	/*
	 * Purge the variables 
	 */
	hpstat = hp_purge (HP_FLAG_VARIABLE_NAME, HP_PURGE_VAR);
	HP_CHECK_HPSTAT (hpstat);
	hpstat = hp_purge (HP_FLAG_RESTORE_NAME, HP_PURGE_VAR);
	HP_CHECK_HPSTAT (hpstat);

	/*
	 * Change to the saved working directory 
	 */
	hpstat = hp_chdir (currdir);
	HP_CHECK_HPSTAT (hpstat);

	return HP_OK;
}

void
hp_cancel_transfer ()
{
	k_cancel_transfer = K_XFER_CANCEL;
}
