
/*
 * HPTalx
 * A PC<->HP48/49 communication program for Linux
 *
 * (C) 1999, 2000 by Bruno Barberi Gnecco
 * (C) 2005 by Bruno Barberi Gnecco, Rafael Ostertag
 *
 * $Id: kermit.c,v 1.3 2007/09/29 21:21:01 guengel Exp $
 */

/*
 * Copyright note:
 * 
 * This program is free software; you can
 * redistribute it and/or modify it under the
 * terms of the GNU General Public License as
 * published by the Free Software Foundation;
 * either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that
 * it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU
 * General Public License along with this
 * program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "hptalx.h"
#include "kermit.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <pty.h>
#include <termios.h>
#include <signal.h>
#include <errno.h>

/*
 * Global variable that holds the last Kermit status 
 */
k_status            k_errno;

/*
 * The global variable indicating whether the current transfer should be
 * canceled 
 */
k_cancel_transfer_switches k_cancel_transfer;

/*
 * The master file descriptor used for communicating with Kermit 
 */
int                 g_master_fld = -1;

/*
 * The pid of the Kermit Process 
 */
pid_t               g_child = -1;

/*
 * Structure holding the Kermit settings 
 */
k_params            g_kermit_params;

/*
 * Holds the transfer mode switches
 *
 * k_transfer_file_switches[0] Automatic transfer mode.
 * k_transfer_file_switches[1] Binary mode. k_transfer_file_switches[2] Text
 * mode.
 */
const char         *k_transfer_file_switches[] = { "", "/BINARY", "/TEXT" };

static k_status     _k_parent (const k_params * params);

/*--------------------------------------------------------------------------*/

static              k_status
_k_parent (const k_params * params)
{
	pid_t pid;
	int status;

	if (fcntl (g_master_fld, F_SETFL, O_NONBLOCK) == -1)
		return K_FCNTL;

	memcpy (&g_kermit_params, params, sizeof (k_params));

	/*
	 * Get rid of startup banner
	 */
	k_errno = k_discard_output ();

	/*
	 * _k_parent() gets called only and immediately after forking child.
	 * All file descriptor parameters are set and the first read took place.
	 * If child exited already with an error kermit might not have been found.
	 */
	pid = waitpid(g_child, &status, WNOHANG);
	if (pid && WIFEXITED(status) && WEXITSTATUS(status) == 2) {
		return k_errno = K_EXECKERMIT;
	}

	k_errno = k_send_command (K_CMD_SET_PROMPT);
	k_errno = k_discard_output ();

	k_errno = k_send_command (K_CMD_QUIET_ON);
	k_errno = k_discard_output ();

	k_errno = k_send_command (K_CMD_HINTS_OFF);
	k_errno = k_discard_output ();

	k_errno = k_send_command (K_CMD_LINE, params->port);
	k_errno = k_discard_output ();

	k_errno = k_send_command (K_CMD_SPEED, params->speed);
	k_errno = k_discard_output ();

	k_errno = k_send_command (K_CMD_FILE_DISPLAY_CRT);
	k_errno = k_discard_output ();

	return k_errno = K_OK;
}

k_status
k_connect (const k_params * params)
{
	struct termios      term_cap;
	struct winsize      size;

	if (g_child != -1) {
		return k_errno = K_CONNECTED;
	}

	/*
	 * Set terminal capabilities
	 */
	term_cap.c_iflag = BRKINT | INLCR;
	term_cap.c_oflag = OCRNL | OFILL | OFDEL | ONLRET | ONOCR;
	term_cap.c_cflag = CLOCAL;
	term_cap.c_lflag = 0;

	/*
	 * Window size of terminal
	 */
	size.ws_row = 25;
	size.ws_col = 80;
	size.ws_xpixel = 0;
	size.ws_ypixel = 0;

	g_child = forkpty (&g_master_fld, NULL, &term_cap, &size);
	if (g_child == -1)
		return k_errno = K_FORKPTY;
	if (g_child == 0) {			/* child */
		if (execlp ("kermit", "kermit", "-z", "-Y", NULL) < 0)
			exit (2);
		exit (0);
	} else {					/* parent */
		tcsetattr (g_master_fld, TCSANOW, &term_cap);
		return _k_parent (params);
	}
}

k_status
k_interrupt (void)
{

	K_CHECKSTATE;
	kill (g_child, SIGINT);
	k_errno = k_discard_output ();
	return k_errno;
}

k_status
k_cancel_action (void)
{
	char                z = 'X';

	K_CHECKSTATE;
	k_errno = k_send_command ("%c", z);
	if (k_errno != K_OK)
		return k_errno;
	k_discard_output ();
	return k_errno;
}

k_status
k_is_connected ()
{
	if (g_master_fld > 0 && g_child > 0)
		return k_errno = K_CONNECTED;
	else
		return k_errno = K_NOTCONNECTED;
}

k_status
k_quit (k_quit_mode m)
{

	K_CHECKSTATE;

	/*
	 * A preventive interrupt 
	 */
	k_interrupt ();

	if (m == K_FULL_QUIT) {
		k_errno = k_send_command (K_CMD_FINISH);
		k_discard_output ();
	}

	k_errno = k_send_command (K_CMD_CLOSE);
	k_discard_output ();

	k_errno = k_send_command (K_CMD_QUIT);
	if (k_errno == K_OK) {
		g_master_fld = -1;
		g_child = -1;
	}

	/*
	 * Wait for the child to terminate 
	 */
	wait (NULL);

	return k_errno;
}

k_status
k_discard_output (void)
{
	char               *nothing;
	size_t              wdc;
	k_status            tmp_stat;

	nothing = NULL;

	tmp_stat = k_wait_ready (&nothing, &wdc);

	k_free_bulk_output (&nothing);

	return k_errno = tmp_stat;
}

k_status
k_wait_ready (char **o, size_t * size)
{
	fd_set              rset;
	struct timeval      tv;
	int                 ret_val;
	int                 br;
	char                buff[K_CHUNKSIZE];

	if (*o != NULL || o == NULL || size == NULL)
		return k_errno = K_INVARG;

	K_CHECKSTATE;

	/*
	 * Set the size argument *s to zero, so we
	 * don't give the caller a false impression
	 * if we fail 
	 */
	*size = 0;

	tcdrain (g_master_fld);

	FD_ZERO (&rset);
	FD_SET (g_master_fld, &rset);
	tv.tv_sec = g_kermit_params.timeout;
	tv.tv_usec = 0;
	ret_val = select (FD_SETSIZE, &rset, NULL, NULL, &tv);
	if (ret_val == 0) {
		return k_errno = K_TIMEOUT;
	}

	*size = 0;
	do {
		/*
		 * This is because select() modifies tv, so
		 * we reset to our values. 
		 */
		tv.tv_sec = g_kermit_params.timeout;
		tv.tv_usec = 0;

		br = read (g_master_fld, buff, K_CHUNKSIZE);

		/*
		 * If br<0 then probably we got no data
		 * in g_master_fld and we wait until data 
		 * has arrived 
		 */
		if (br > 0) {

			*size += br;
			
			if (*size == br) {
				*o = (char *) malloc (*size);
			} else {
				*o = (char *) realloc (*o, *size);
			}
			
			if (*o == NULL) {
				return k_errno = K_MALLOC;
			}

			memcpy ((void *) (*o + (*size - br)), (void *) buff, br);

		} else if (br == 0 || errno == EAGAIN) {
			/*
			 * Wait nicely for new data
			 */
			ret_val = select (FD_SETSIZE, &rset, NULL, NULL, &tv);
			if (ret_val < 1) {
				return k_errno = K_TIMEOUT;
			}

		} else {
			return k_errno = K_READ;
		}

		/*
		 * Probably superfluous code 
		 */
		tv.tv_sec = 0;
		tv.tv_usec = 0;

	}
	while ((*o != 0 && strstr (*o, K_KERMIT_PROMPT) == NULL)
		   || (select (FD_SETSIZE, &rset, NULL, NULL, &tv) != 0));
	return k_errno = K_OK;
}

k_status
k_read_file_transfer (k_transfer_callback callback, void *data)
{

	FILE               *f;
	fd_set              rset;
	struct timeval      tv;
	int                 ret_val;
	char                buff[K_BUFFERSIZE];
	unsigned int        cps;
	unsigned int        bytes;
	unsigned int        p;

	K_CHECKSTATE;

	if (callback == NULL)
		return k_errno = K_INVARG;

	tcdrain (g_master_fld);

	FD_ZERO (&rset);
	FD_SET (g_master_fld, &rset);
	tv.tv_sec = g_kermit_params.timeout;
	tv.tv_usec = 0;
	ret_val = select (FD_SETSIZE, &rset, NULL, NULL, &tv);
	if (ret_val == 0) {
		return k_errno = K_TIMEOUT;
	}
	/*
	 * This is because select() modifies tv, so
	 * we reset to our values. 
	 */
	tv.tv_sec = g_kermit_params.timeout;
	tv.tv_usec = 0;

	/*
	 * Open a stream, so we can read from it 
	 */
	f = fdopen (g_master_fld, "r");
	if (f == NULL) {
		return k_errno = K_STREAM;
	}

	/*
	 * Initialize the flag indicating whether we should cancel the transfer 
	 */
	k_cancel_transfer = K_XFER_PROCEED;

	do {

		if (k_cancel_transfer == K_XFER_CANCEL) {
			k_interrupt ();
			/*
			 * Reset the flage 
			 */
			k_cancel_transfer = K_XFER_PROCEED;
			return k_errno = K_CANCEL;
		}

		if (fgets (buff, K_BUFFERSIZE, f) != NULL) {
			p = bytes = cps = 0;
			sscanf (buff, "%u%u%%%u", &bytes, &p, &cps);
			if (bytes != 0 || cps != 0 || p != 0) {
				callback (&bytes, &p, &cps, data);
			}

		} else {
			/*
			 * Wait nicely for new data
			 */
			ret_val = select (FD_SETSIZE, &rset, NULL, NULL, &tv);
			if (ret_val == 0) {
				return k_errno = K_TIMEOUT;
			}
			/*
			 * This is because select() modifies
			 * tv, so we reset to our values. 
			 */
			tv.tv_sec = g_kermit_params.timeout;
			tv.tv_usec = 0;
		}

	}
	while (strstr (buff, K_KERMIT_PROMPT) == NULL);

	return k_errno = K_OK;
}

k_status
k_read_server_transfer (k_transfer_callback callback, void *data)
{

	FILE               *f;
	fd_set              rset;
	struct timeval      tv;
	int                 ret_val;
	char                buff[K_BUFFERSIZE];
	unsigned int        cps;
	unsigned int        bytes;
	unsigned int        p;

	K_CHECKSTATE;

	if (callback == NULL)
		return k_errno = K_INVARG;

	tcdrain (g_master_fld);

	FD_ZERO (&rset);
	FD_SET (g_master_fld, &rset);
	tv.tv_sec = g_kermit_params.timeout;
	tv.tv_usec = 0;
	ret_val = select (FD_SETSIZE, &rset, NULL, NULL, &tv);
	if (ret_val == 0) {
		return k_errno = K_TIMEOUT;
	}

	/*
	 * This is because select() modifies tv, so
	 * we reset to our values. 
	 */
	tv.tv_sec = g_kermit_params.timeout;
	tv.tv_usec = 0;

	/*
	 * Open a stream, so we can read from it 
	 */
	f = fdopen (g_master_fld, "r");
	if (f == NULL) {
		return k_errno = K_STREAM;
	}

	/*
	 * Initialize the flag indicating whether we should cancel the transfer 
	 */
	k_cancel_transfer = K_XFER_PROCEED;

	do {

		if (k_cancel_transfer == K_XFER_CANCEL) {
			/*
			 * Send an error packet 
			 */
			k_send_command ("E");
			sleep (1);

			/*
			 * Now terminate 
			 */
			k_interrupt ();
			/*
			 * Reset the flage 
			 */
			k_cancel_transfer = K_XFER_PROCEED;
			return k_errno = K_CANCEL;
		}

		if (fgets (buff, K_BUFFERSIZE, f) != NULL) {
			p = bytes = cps = 0;
			sscanf (buff, "%u%u%%%u", &bytes, &p, &cps);
			if (bytes != 0 || cps != 0 || p != 0) {
				callback (&bytes, &p, &cps, data);
			}

		} else {
			/*
			 * Wait nicely for new data
			 */
			ret_val = select (FD_SETSIZE, &rset, NULL, NULL, &tv);
			if (ret_val == 0) {
				return k_errno = K_TIMEOUT;
			}

			/*
			 * This is because select() modifies
			 * tv, so we reset to our values. 
			 */
			tv.tv_sec = g_kermit_params.timeout;
			tv.tv_usec = 0;
		}

	}
	while (strstr (buff, K_TRANSFER_OK) == NULL);

	return k_errno = K_OK;
}

k_status
k_free_bulk_output (char **o)
{
	if (*o == NULL)
		return k_errno = K_INVARG;
	if (o == NULL)
		return k_errno = K_INVARG;
	free (*o);
	*o = NULL;
	return k_errno = K_OK;
}

k_status
k_send_remote_command (const char *fmt, ...)
{
	char               *buf;
	char               *buf_exp;
	va_list             ap;
	int                 ret_val;

	buf = (char *) malloc (K_BUFFERSIZE);
	if (buf == NULL)
		return k_errno = K_MALLOC;

	buf_exp = (char *) malloc (K_BUFFERSIZE);
	if (buf_exp == NULL) {
		free (buf);
		return k_errno = K_MALLOC;
	}

	memset ((void *) buf, 0, K_BUFFERSIZE);
	memset ((void *) buf_exp, 0, K_BUFFERSIZE);

	va_start (ap, fmt);
	ret_val = vsnprintf (buf, K_BUFFERSIZE, fmt, ap);
	va_end (ap);
	if (ret_val < 0) {
		free (buf);
		free (buf_exp);
		return k_errno = K_WRITE;
	}

	/*
	 * Simply prepend "REMOTE HOST " in front of fmt 
	 */
	if (K_BUFFERSIZE < (strlen (K_CMD_REMOTE_HOST) + 1))
		return K_BUFFER;
	strncpy (buf_exp, K_CMD_REMOTE_HOST, strlen (K_CMD_REMOTE_HOST));
	if (K_BUFFERSIZE < (strlen (buf) + 1))
		return K_BUFFER;
	strncat (buf_exp, buf, strlen (buf));
	free (buf);

	k_errno = k_send_command ("%s", buf_exp);

	free (buf_exp);

	return k_errno;
}

k_status
k_send_command (const char *fmt, ...)
{
	va_list             ap;
	char                buf[K_BUFFERSIZE];
	int                 ret_val;
	ssize_t             bw;
	fd_set              fdset;
	struct timeval      tv;

	K_CHECKSTATE;

	memset ((void *) buf, 0, K_BUFFERSIZE);
	va_start (ap, fmt);
	ret_val = vsnprintf (buf, K_BUFFERSIZE, fmt, ap);
	va_end (ap);
	if (ret_val < 0) {
		return k_errno = K_WRITE;
	}

	/*
	 * Add trailling the newline in order to make Kermit recognize it as
	 * input 
	 */
	if (K_BUFFERSIZE < (strlen (buf) + strlen (K_STR_NEWLINE) + 1))
		return K_BUFFER;
	strncat (buf, K_STR_NEWLINE, strlen (K_STR_NEWLINE));

	FD_ZERO (&fdset);
	FD_SET (g_master_fld, &fdset);

	tv.tv_sec = g_kermit_params.timeout;
	tv.tv_usec = 0;
	ret_val = select (FD_SETSIZE, NULL, &fdset, NULL, &tv);
	if (ret_val == 0) {
		return k_errno = K_TIMEOUT;
	}
	if (ret_val == -1) {
		return k_errno = K_SELECT;
	}

	bw = write (g_master_fld, (void *) buf, strlen (buf));
	if (bw != strlen (buf)) {
		return k_errno = K_WRITE;
	}

	tcdrain (g_master_fld);

	return k_errno = K_OK;
}
