
/*
 * HPTalx
 * A PC<->HP48/49 communication program for Linux
 *
 * (C) 1999, 2000 by Bruno Barberi Gnecco
 * (C) 2005 by Rafael Ostertag
 *
 * rev 1.1.0:   File split from io.c
 *      Added kermit/xmodem support
 *      Added wait_connect field
 *
 * $Id: setup.c,v 1.3 2007/09/29 21:21:01 guengel Exp $
 *
 *
 * Copyright note:
 *
 * This program is free software; you can
 * redistribute it and/or modify it under the
 * terms of the GNU General Public License as
 * published by the Free Software Foundation;
 * either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that
 * it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU
 * General Public License along with this
 * program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "hptalx.h"
#include "setup.h"
#include "gui_utils.h"

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>

/*
 * The available port speeds
 */
const char         *speed_list[] =
    { "9600", "19200", "38400", "57600", "115200", "230400", "460800", NULL };

/*
 * Used when callback _set_connection_type is
 * called
 */
struct _conntype_callback_args
{
    CONNECTION_TYPE     t;
    GtkWidget          *box_serial;
    GtkWidget          *box_ttyusb;
};
typedef struct _conntype_callback_args conntype_callback_args;

#define P_CONNTYPE_CALLBACK_ARGS(a) ((conntype_callback_args*)a)

static void         _populate_speed_menu (GtkWidget ** options);
static int          _get_item_index_speed_menu (const char *speed);
static void         _set_text (GtkWidget * w, gpointer data);
static void         _set_com_port (GtkWidget * w, gpointer data);
static void         _set_ttyusb_port (GtkWidget * w, gpointer data);
static void         _set_port_speed (GtkWidget * w, gpointer data);
static void         _set_wait_connect (GtkWidget * w, gpointer data);
static void         _set_toggle (GtkWidget * w, gpointer data);
static void         _get_pref_filename (char **filename);
static GtkWidget   *_create_setserialport_box (GtkWidget * box);
static GtkWidget   *_create_setttyusbport_box (GtkWidget * box);
static void         _set_connection_type (GtkWidget * w, gpointer data);
static void         _save_prefs (GtkWidget * w, gpointer data);
static int          _xml_string_to_int (const xmlChar * xc);

/*--------------------------------------------------------------------------*/

static void
_populate_speed_menu (GtkWidget ** options)
{
    GtkWidget          *menu_item;
    const char        **l = speed_list;

    *options = gtk_menu_new ();

    while (*l != NULL) {
	/*
	 * Create a new menu item for the Option Menu 
	 */
	menu_item = gtk_menu_item_new_with_label ((char *) *l);
	/*
	 * Connect the signal handler. The Port Speed string *l is passed as
	 * argument, so the signal handler can get the speed selected and
	 * convert it into a number 
	 */
	g_signal_connect (GTK_OBJECT (menu_item), "activate",
			  G_CALLBACK (_set_port_speed), (gpointer) * l);
	/*
	 * Finaly append the menu item to the menu 
	 */
	gtk_menu_append (GTK_MENU (*options), menu_item);
	l++;
    }
}

/*
 * Given a port speed as string, it searches the speed_list an returns the
 * position of *speed in the speed_list array as the index of the menu entry
 * holding that port speed 
 */
static int
_get_item_index_speed_menu (const char *speed)
{
    int                 i;
    const char        **l = speed_list;

    if (speed == NULL)
	return -1;

    i = 0;
    while (*l != NULL) {
	if (strcmp (*l, speed) == 0)
	    break;
	i++;
	l++;
    }
    return i;
}

static void
_set_text (GtkWidget * w, gpointer data)
{
    if (preferences.editor)
	free (preferences.editor);
    preferences.editor = strdup (gtk_entry_get_text (data));
}

static void
_set_com_port (GtkWidget * w, gpointer data)
{
    preferences.port = GPOINTER_TO_INT (data);
}

static void
_set_ttyusb_port (GtkWidget * w, gpointer data)
{
    preferences.port =
	gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (w)) + 100;
}

/*
 * data holds the Port Speed string which is converted to a number 
 */
static void
_set_port_speed (GtkWidget * w, gpointer data)
{
    preferences.speed = atoi ((char *) data);
}

static void
_set_wait_connect (GtkWidget * w, gpointer data)
{
    preferences.wait_connect =
	gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (data));

}

static void
_set_toggle (GtkWidget * w, gpointer data)
{
    if (!strcmp (data, "delete_confirm"))
	preferences.delete_confirm = GTK_TOGGLE_BUTTON (w)->active;
    else if (!strcmp (data, "start_connected"))
	preferences.start_connected = GTK_TOGGLE_BUTTON (w)->active;
    else
	message_box ("!", "Well, this message means that either\n"
		     "something has gone very wrong or there's a bug\n"
		     "in the _set_toggle() function...");
}

static void
_get_pref_filename (char **filename)
{
    char               *dir;
    char               *t;

    dir = strdup (getenv ("HOME"));
    t = strtok (dir, "=");
    while (isspace (*t))
	t++;
    *filename = (char *) malloc (strlen (t) + 9);
    strcpy (*filename, t);
    strcat (*filename, "/.hptalx");
    free (dir);
}

static GtkWidget   *
_create_setserialport_box (GtkWidget * box)
{
    GtkWidget          *box_serial;
    GtkWidget          *label;
    GtkWidget          *vbox;
    GtkWidget          *button;

    /*
     * Create the 'Set COM port' hbox
     */
    box_serial = gtk_hbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box_serial), 10);
    gtk_box_pack_start (GTK_BOX (box), box_serial, TRUE, TRUE, 0);

    /*
     * Radio buttons to set serial port 
     */
    label = gtk_label_new ("Set COM port:");
    gtk_box_pack_start (GTK_BOX (box_serial), label, TRUE, FALSE, 0);

    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);
    gtk_box_pack_start (GTK_BOX (box_serial), vbox, TRUE, TRUE, 0);

    /*
     * Com 1 
     */
    button = gtk_radio_button_new_with_label (NULL, "COM 1");
    g_signal_connect (GTK_OBJECT (button), "clicked",
		      G_CALLBACK (_set_com_port), GINT_TO_POINTER (0));
    gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, TRUE, 0);
    if (preferences.port == 0)
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);

    /*
     * Com 2 
     */
    button =
	gtk_radio_button_new_with_label (gtk_radio_button_group
					 (GTK_RADIO_BUTTON (button)),
					 "COM 2");
    g_signal_connect (GTK_OBJECT (button), "clicked",
		      G_CALLBACK (_set_com_port), GINT_TO_POINTER (1));
    gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, TRUE, 0);
    if (preferences.port == 1)
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);

    /*
     * Com 3 
     */
    button =
	gtk_radio_button_new_with_label (gtk_radio_button_group
					 (GTK_RADIO_BUTTON (button)),
					 "COM 3");
    g_signal_connect (GTK_OBJECT (button), "clicked",
		      G_CALLBACK (_set_com_port), GINT_TO_POINTER (2));
    gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, TRUE, 0);
    if (preferences.port == 2)
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);

    /*
     * Com 4 
     */
    button =
	gtk_radio_button_new_with_label (gtk_radio_button_group
					 (GTK_RADIO_BUTTON (button)),
					 "COM 4");
    g_signal_connect (GTK_OBJECT (button), "clicked",
		      G_CALLBACK (_set_com_port), GINT_TO_POINTER (3));
    gtk_box_pack_start (GTK_BOX (vbox), button, TRUE, TRUE, 0);
    if (preferences.port == 3)
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);

    return box_serial;
}

static GtkWidget   *
_create_setttyusbport_box (GtkWidget * box)
{
    GtkWidget          *box_ttyusb;
    GtkWidget          *label;
    GtkWidget          *button;

    GtkAdjustment      *interval;

    /*
     * Create the 'Set ttyUSB port' hbox
     */
    box_ttyusb = gtk_hbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box_ttyusb), 10);
    gtk_box_pack_start (GTK_BOX (box), box_ttyusb, TRUE, TRUE, 0);

    /*
     * The label 
     */
    label = gtk_label_new ("Set ttyUSB port:");
    gtk_box_pack_start (GTK_BOX (box_ttyusb), label, TRUE, FALSE, 0);

    /*
     * The spin button for choosing the port number 
     */
    interval = GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 20, 1, 1, 0));
    button = gtk_spin_button_new (interval, 1, 0);
    g_signal_connect (GTK_OBJECT (button), "changed",
		      G_CALLBACK (_set_ttyusb_port), NULL);
    gtk_box_pack_start (GTK_BOX (box_ttyusb), button, TRUE, FALSE, 0);

    if (preferences.port > 99)
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (button),
				   preferences.port - 100);

    return box_ttyusb;
}

static void
_set_connection_type (GtkWidget * w, gpointer data)
{

    switch (P_CONNTYPE_CALLBACK_ARGS (data)->t) {
    case SERIAL:
	/*
	 * Hide the ttyUSB box 
	 */
	if (P_CONNTYPE_CALLBACK_ARGS (data)->box_ttyusb != NULL)
	    gtk_widget_hide (P_CONNTYPE_CALLBACK_ARGS (data)->box_ttyusb);
	/*
	 * Show serial box 
	 */
	if (P_CONNTYPE_CALLBACK_ARGS (data)->box_serial != NULL)
	    gtk_widget_show (P_CONNTYPE_CALLBACK_ARGS (data)->box_serial);

	/*
	 * Make serial ports default 
	 */
	preferences.port = 0;
	break;
    case TTYUSB:
	/*
	 * Hide the serial box 
	 */
	if (P_CONNTYPE_CALLBACK_ARGS (data)->box_serial != NULL)
	    gtk_widget_hide (P_CONNTYPE_CALLBACK_ARGS (data)->box_serial);
	/*
	 * Create ttyUSB box only when not already created 
	 */
	if (P_CONNTYPE_CALLBACK_ARGS (data)->box_ttyusb != NULL)
	    gtk_widget_show (P_CONNTYPE_CALLBACK_ARGS (data)->box_ttyusb);

	/*
	 * Make ttyUSB ports default 
	 */
	preferences.port = 100;
	break;
    }

}

static void
_save_prefs (GtkWidget * w, gpointer data)
{

    xmlDocPtr           xml_doc;
    xmlTextWriterPtr    xml_writer;
    char               *filename;

    xml_writer = xmlNewTextWriterDoc (&xml_doc, 0);
    xmlTextWriterSetIndent (xml_writer, 4);
    xmlTextWriterStartDocument (xml_writer, NULL, NULL, NULL);
    /*
     * Create the toplevel tag <hptalx> 
     */
    xmlTextWriterStartElement (xml_writer,
			       (const xmlChar *) CFG_ROOT_ELEMENT);

    /*
     * Save the port number 
     */
    xmlTextWriterWriteFormatElement (xml_writer, (const xmlChar *) CFG_PORT,
				     "%d", preferences.port);
    /*
     * Save the port speed 
     */
    xmlTextWriterWriteFormatElement (xml_writer,
				     (const xmlChar *) CFG_SPEED, "%d",
				     preferences.speed);
    /*
     * Save the delete confirm flag 
     */
    xmlTextWriterWriteFormatElement (xml_writer, (const xmlChar *)
				     CFG_DELETECONFIRM, "%d",
				     preferences.delete_confirm);
    /*
     * Save the start connected flag 
     */
    xmlTextWriterWriteFormatElement (xml_writer, (const xmlChar *)
				     CFG_STARTCONNECTED, "%d",
				     preferences.start_connected);
    /*
     * Save the wait connect time 
     */
    xmlTextWriterWriteFormatElement (xml_writer, (const xmlChar *)
				     CFG_WAITCONNECT, "%d",
				     preferences.wait_connect);
    /*
     * Save the editor command 
     */
    xmlTextWriterWriteFormatElement (xml_writer,
				     (const xmlChar *) CFG_EDITOR, "%s",
				     preferences.editor);

    /*
     * End the toplevel tag <hptalx> 
     */
    xmlTextWriterEndElement (xml_writer);

    /*
     * End the document 
     */
    xmlTextWriterEndDocument (xml_writer);

    /*
     * Flush the writer 
     */
    xmlTextWriterFlush (xml_writer);

    /*
     * Save the document 
     */
    _get_pref_filename (&filename);
    xmlSaveFormatFile (filename, xml_doc, 0);

    /*
     * Free the document 
     */
    xmlFreeDoc (xml_doc);

    /*
     * Free the writer 
     */
    xmlFreeTextWriter (xml_writer);

    free (filename);
}

static int
_xml_string_to_int (const xmlChar * xc)
{
    int                 i;

    sscanf ((const char *) xc, "%i", &i);
    return i;
}

void
load_prefs (void)
{
    xmlDocPtr           cfg;
    xmlNodePtr          xml_root;
    xmlNodePtr          xml_childnode;
    xmlChar            *xml_str;
    char               *filename;

    /*
     * Set the defaults, in case no config file exists yet 
     */
    preferences.port = 0;
    preferences.speed = 9600;
    preferences.delete_confirm = TRUE;
    preferences.start_connected = FALSE;
    preferences.wait_connect = 15;
    preferences.editor = strdup ("gedit");

    /*
     * Open the config file 
     */
    _get_pref_filename (&filename);
    cfg = xmlParseFile (filename);
    if (cfg == NULL) {
	status_push ("Cannot load config file");
	free (filename);
	return;
    }

    xml_root = xmlDocGetRootElement (cfg);
    if (xml_root == NULL) {
	status_push ("Empty config file");
	goto THEEND;
    }

    /*
     * Make sure, it's a hptalx config file 
     */
    if (xmlStrcmp (xml_root->name, (const xmlChar *) CFG_ROOT_ELEMENT)) {
	status_push ("Configuration file of the wrong type");
	goto THEEND;
    }

    /*
     * Now, get all the settings 
     */
    xml_childnode = xml_root->xmlChildrenNode;
    while (xml_childnode != NULL) {
	if (xmlStrEqual (xml_childnode->name, (const xmlChar *) CFG_PORT)) {
	    xml_str =
		xmlNodeListGetString (cfg, xml_childnode->xmlChildrenNode, 1);
	    preferences.port = _xml_string_to_int (xml_str);
	    xmlFree (xml_str);
	    goto THENEXTPLEASE;
	}
	if (xmlStrEqual (xml_childnode->name, (const xmlChar *) CFG_SPEED)) {
	    xml_str =
		xmlNodeListGetString (cfg, xml_childnode->xmlChildrenNode, 1);
	    preferences.speed = _xml_string_to_int (xml_str);
	    xmlFree (xml_str);
	    goto THENEXTPLEASE;
	}
	if (xmlStrEqual
	    (xml_childnode->name, (const xmlChar *) CFG_DELETECONFIRM)) {
	    xml_str =
		xmlNodeListGetString (cfg, xml_childnode->xmlChildrenNode, 1);
	    preferences.delete_confirm = _xml_string_to_int (xml_str);
	    xmlFree (xml_str);
	    goto THENEXTPLEASE;
	}
	if (xmlStrEqual
	    (xml_childnode->name, (const xmlChar *) CFG_STARTCONNECTED)) {
	    xml_str =
		xmlNodeListGetString (cfg, xml_childnode->xmlChildrenNode, 1);
	    preferences.start_connected = _xml_string_to_int (xml_str);
	    xmlFree (xml_str);
	    goto THENEXTPLEASE;
	}
	if (xmlStrEqual
	    (xml_childnode->name, (const xmlChar *) CFG_WAITCONNECT)) {
	    xml_str =
		xmlNodeListGetString (cfg, xml_childnode->xmlChildrenNode, 1);
	    preferences.wait_connect = _xml_string_to_int (xml_str);
	    xmlFree (xml_str);
	    goto THENEXTPLEASE;
	}
	if (xmlStrEqual (xml_childnode->name, (const xmlChar *) CFG_EDITOR)) {
	    xml_str =
		xmlNodeListGetString (cfg, xml_childnode->xmlChildrenNode, 1);
	    /*
	     * Free the memory previously was allocated by preventive filling 
	     * in the values 
	     */
	    free (preferences.editor);
	    preferences.editor = strdup ((const char *) xml_str);
	    xmlFree (xml_str);
	    goto THENEXTPLEASE;
	}
      THENEXTPLEASE:
	xml_childnode = xml_childnode->next;
    }

  THEEND:
    xmlFreeDoc (cfg);
    free (filename);
}

void
setup (GtkWidget * w, gpointer data)
{
    GtkWidget          *dialog;
    GtkWidget          *label;
    GtkWidget          *hbox;
    GtkWidget          *vbox;
    GtkWidget          *frame;
    GtkWidget          *button;
    GtkWidget          *button_serial;
    GtkWidget          *button_ttyusb;
    GtkWidget          *sep;
    GtkWidget          *entry;
    GtkWidget          *port_speed_widget;
    GtkWidget          *wait_but;
    GtkWidget          *port_speed_menu;
    GtkAdjustment      *wait_adj;
    char               *port_speed;
    conntype_callback_args conntype_serial;
    conntype_callback_args conntype_ttyusb;

    dialog = gtk_dialog_new ();

    /*
     * Make the dialog modal
     */
    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
    gtk_window_set_title (GTK_WINDOW (dialog), "Setup");
    gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
    gtk_widget_set_usize (dialog, 0, 0);

    /*
     * Frame for connection settings 
     */
    frame = gtk_frame_new ("Connection Settings");
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), frame, TRUE,
			FALSE, 2);

    /*
     * Create the vbox used in the frame create above 
     */
    vbox = gtk_vbox_new (FALSE, 2);
    gtk_container_add (GTK_CONTAINER (frame), vbox);

    /*
     * Box for selection whether ttyS* or ttyUSB* have to be used 
     */
    hbox = gtk_hbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, FALSE, 2);

    /*
     * Label for the Connection type 
     */
    button_serial = gtk_radio_button_new_with_label (NULL, "Serial");
    g_signal_connect (GTK_OBJECT (button_serial), "clicked",
		      G_CALLBACK (_set_connection_type), &conntype_serial);
    gtk_box_pack_start (GTK_BOX (hbox), button_serial, TRUE, FALSE, 0);

    button_ttyusb =
	gtk_radio_button_new_with_label (gtk_radio_button_group
					 (GTK_RADIO_BUTTON (button_serial)),
					 "USB (ttyUSB)");
    g_signal_connect (GTK_OBJECT (button_ttyusb), "clicked",
		      G_CALLBACK (_set_connection_type), &conntype_ttyusb);
    gtk_box_pack_start (GTK_BOX (hbox), button_ttyusb, TRUE, FALSE, 0);

    /*
     * Seperator 
     */
    sep = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (vbox), sep, TRUE, FALSE, 5);

    /*
     * Load arguments for callback for connection type serial. It is
     * loaded here, because the seperator has to be in the right place,
     * and while loading arguments the boxes are created by the call to _create_setserialpor_box() and _create_setttyusbport_box()...
     */

    conntype_serial.box_serial = conntype_ttyusb.box_serial =
	_create_setserialport_box (vbox);
    conntype_serial.box_ttyusb = conntype_ttyusb.box_ttyusb =
	_create_setttyusbport_box (vbox);

    conntype_serial.t = SERIAL;
    conntype_ttyusb.t = TTYUSB;

    /*
     * Seperator 
     */
    sep = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (vbox), sep, TRUE, TRUE, 0);

    /*
     * Port Speed Option Menu
     */
    hbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);

    /*
     * Port Speed Label 
     */
    label = gtk_label_new ("Set COM Speed:");
    gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 3);

    /*
     * Create the Port Speed Option Menu Widget 
     */
    port_speed_widget = gtk_option_menu_new ();
    gtk_box_pack_start (GTK_BOX (hbox), port_speed_widget, TRUE, TRUE, 0);

    /*
     * Create the Menu Widget passed to the Port Speed Option Menu and
     * connect signal handlers
     */
    _populate_speed_menu (&port_speed_menu);
    gtk_option_menu_set_menu (GTK_OPTION_MENU (port_speed_widget),
			      port_speed_menu);

    /*
     * Set port speed according the Preferences in the Port Speed Option Menu
     */
    port_speed = (char *) malloc (MAXPORTSPEEDSIZE);
    sprintf (port_speed, "%d", preferences.speed);
    gtk_option_menu_set_history (GTK_OPTION_MENU (port_speed_widget),
				 _get_item_index_speed_menu (port_speed));
    free (port_speed);

    /*
     * Seperator 
     */
    sep = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (vbox), sep, TRUE, TRUE, 0);

    /*
     * Timeout label 
     */
    hbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);

    label = gtk_label_new ("Timeout:");
    gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 2);

    /*
     * Timeout Adjust
     */
    wait_adj =
	(GtkAdjustment *) gtk_adjustment_new (preferences.wait_connect,
					      0.0, 120.0, 1.0, 5.0, 0.0);
    wait_but = gtk_spin_button_new (wait_adj, 0, 0);
    gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (wait_but), TRUE);
    g_signal_connect (GTK_OBJECT (wait_adj), "value_changed",
		      G_CALLBACK (_set_wait_connect), (gpointer) wait_but);

    gtk_box_pack_start (GTK_BOX (hbox), wait_but, TRUE, TRUE, 0);

    /*
     * Here the frame for the Connection Settings ends and other options are
     * created 
     */

    /*
     * confirm delete? 
     */
    button = gtk_check_button_new_with_label ("Confirm Delete and Remove");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
				  preferences.delete_confirm);
    g_signal_connect (GTK_OBJECT (button), "clicked",
		      G_CALLBACK (_set_toggle), "delete_confirm");
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), button, TRUE,
			TRUE, 0);

    /*
     * start connected? 
     */
    button = gtk_check_button_new_with_label ("Start Connected");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
				  preferences.start_connected);
    g_signal_connect (GTK_OBJECT (button), "clicked",
		      G_CALLBACK (_set_toggle), "start_connected");
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), button, TRUE,
			TRUE, 0);

    /*
     * editor 
     */
    hbox = gtk_hbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox), 10);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE,
			TRUE, 0);

    label = gtk_label_new ("Set default editor:");
    gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, FALSE, 0);

    entry = gtk_entry_new ();
    gtk_entry_set_text (GTK_ENTRY (entry), preferences.editor);
    gtk_box_pack_start (GTK_BOX (hbox), entry, FALSE, FALSE, 10);
    g_signal_connect (GTK_OBJECT (entry), "changed",
		      G_CALLBACK (_set_text), GTK_OBJECT (entry));

    button = gtk_button_new_with_label ("OK");
    gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			       G_CALLBACK (gtk_main_quit),
			       GTK_OBJECT (dialog));
    gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			       G_CALLBACK (gtk_widget_destroy),
			       GTK_OBJECT (dialog));
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), button,
			TRUE, TRUE, 0);
    GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default (button);

    button = gtk_button_new_with_label ("Save");
    gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			       G_CALLBACK (_save_prefs), GTK_OBJECT (dialog));
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
			button, TRUE, TRUE, 0);

    gtk_widget_show_all (dialog);

    /*
     * Get the connection type and hide boxes accordingly
     *
     *
     * Ports above 100 are meant for ttyUSB. To get the right port
     * number, one has to subtract 100 to get the port the user
     * intended.
     */
    if (preferences.port > 99) {
	/*
	 * ttyUSB type 
	 */
	_set_connection_type (dialog, &conntype_ttyusb);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button_ttyusb),
				      TRUE);
    } else {
	/*
	 * serial type 
	 */
	_set_connection_type (dialog, &conntype_serial);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button_serial),
				      TRUE);
    }
    gtk_main ();
}
